/*********************************************************************************

 ** Universidade Nova de Lisboa - Faculdade de Ciências e Tecnologia

 **

 ** Protocolo XORi - flatbeacon.cc

 **

 ** Autores:

 ** André Garrido

 ** Rodolfo Oliveira

 **

 ** 1 de Março de 2010

** **********************************************************************************/






#include <address.h>
#include <map>
#include "flatbeacon.h"
#include "cmu-trace.h"




static void flatbeacon_mac_failed_callback(Packet *p, void *arg) {
	((flatbeacon*) arg)->mac_failed(p);
}


///////////////TCL Hooks///////////////////


int hdr_flatbeacon_pkt::offset_;
int hdr_flatbeacon_query_pkt::offset_;
static class FlatouringHeaderClass: public PacketHeaderClass {
public:
	FlatouringHeaderClass() :
		PacketHeaderClass("PacketHeader/Flatbeacon", sizeof(hdr_flatbeacon_pkt)) {
		bind_offset(&hdr_flatbeacon_pkt::offset_);
	}
} class_rtProtoflatbeacon_hdr;


static class flatbeaconClass: public TclClass {
public:
	flatbeaconClass() :
		TclClass("Agent/Flatbeacon") {
	}
	TclObject* create(int argc, const char* const * argv) {
		assert(argc == 5);
		return (new flatbeacon((nsaddr_t) Address::instance().str2addr(argv[4])));
	}
} class_rtProtoflatbeacon;


///Sends a flabeacon message and reschedules the flatbeacon timer.

void flatbeacon_PktTimer::expire(Event* e) {
	agent_->send_flatbeacon_pkt();

	agent_->reset_flatbeacon_pkt_timer();

}


//Sends a query and reschedules the QUERY timer

void flatbeacon_PktTimer_query::expire(Event* e) {


	agent_->send_query();

	agent_->reset_timer_query();
}


//Sets the state of BGL to 0 and reschedules the timer of BGL.

void flatbeacon_PktTimer_bgl_timeout::expire(Event* e){

	agent_->is_bgl_queried= 0;
	agent_->reset_timer_query();
}

//Sends the routing table to the BGL and reschedules the timer.
void flatbeacon_PktTimer_send_bgl_timeout::expire(Event* e){

	agent_->send_bgl();
	agent_->reset_timer_send_bgl();
}


void flatbeacon::mac_failed(Packet* p)
{
	struct hdr_ip* ih    = HDR_IP(p);
	struct hdr_cmn* ch = HDR_CMN(p);

	debug("%f: Node %d MAC layer cannot send a packet to node %d\n",CURRENT_TIME,ra_addr(),ch->next_hop());

	drop(p, DROP_RTR_MAC_CALLBACK);

}


/// Creates necessary timers, binds TCL-available variables and do
/// some more initializations.

flatbeacon::flatbeacon(nsaddr_t id) :
	Agent(PT_FLATBEACON), pkt_timer_(this), pkt_timer_qry_(this), pkt_timer_bgl_timeout_(this), pkt_timer_send_bgl_timeout_(this) {

	bind_bool("accessible_var_", &accessible_var_);
	bind_bool("use_mac_", &use_mac_);


	ra_addr_ = id;
	seq_num_ = 0;
	K_=0;
	codebit_=0;
	tab_bc = (Beacons_tab*) new Beacons_tab(this);
	tab_bc->address = ra_addr_;
	is_bgl_queried= 0;
	neigh_table_= (flatbeacon_neigb_table*)new flatbeacon_neigb_table(this,ra_addr_);

	is_online= false;

	BGL = -1;
}

flatbeacon::~flatbeacon() {

	this->pkt_timer_.force_cancel();
	delete tab_bc;

}


/////Interface with TCL interpreter.





int flatbeacon::command(int argc, const char* const * argv) {
	if (argc == 2) {
		//Starts all timers
		if (strcasecmp(argv[1], "start") == 0) {
			pkt_timer_.resched(0.0);

			is_online= true;

			double interval= Random::uniform(3);

			pkt_timer_qry_.resched(interval);

			return TCL_OK;

			//Stops all timers
		}else if(strcasecmp(argv[1], "stop")==0)
		{
			pkt_timer_.force_cancel();
			pkt_timer_qry_.force_cancel();
			pkt_timer_bgl_timeout_.force_cancel();
			pkt_timer_send_bgl_timeout_.force_cancel();

			is_online= false;

			delete tab_bc;

			return TCL_OK;
		} else if (strcasecmp(argv[1], "print_rtable") == 0) {
			if (logtarget_ != 0) {
				sprintf(logtarget_->pt_->buffer(), "P %f _%d_ Routing Table",
						CURRENT_TIME, ra_addr());
				logtarget_->pt_->dump();
				rtable_.print(logtarget_);
			} else {
				fprintf(stdout,
						"%f _%d_ If you want to print this routing table "
							"you must create a trace file in your tcl script",
						CURRENT_TIME, ra_addr());
			}
			return TCL_OK;

			//Print the routing table
		} else if(strcmp(argv[1],"print_neigb")==0){

			neigh_table_->print(ra_addr_,codebit_);
			return TCL_OK;

			//Print the cache table
		} else if(strcmp(argv[1],"print_cache")==0){

			cache_->print_query_cache();
			return TCL_OK;
		}
	} else if (argc == 3) {
		// Obtains corresponding dmux to carry packets to upper layers
		if (strcmp(argv[1], "port-dmux") == 0) {
			dmux_ = (Classifier*) TclObject::lookup(argv[2]);
			if (dmux_ == 0) {
				fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__,
						argv[1], argv[2]);
				return TCL_ERROR;
			}
			return TCL_OK;
		}
		// Obtains corresponding tracer
		else if (strcmp(argv[1], "log-target") == 0 || strcmp(argv[1],
				"tracetarget") == 0) {
			logtarget_ = (Trace*) TclObject::lookup(argv[2]);
			if (logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;

			//Set the number of bits
		} else if(strcmp(argv[1], "codebit") == 0)
		{
			codebit_= (atoi(argv[2]));
			return TCL_OK;

			//Value of K
		} else if(strcmp(argv[1], "K") == 0)
		{
			K_=atoi(argv[2]);
			return TCL_OK;
		} else if(strcmp(argv[1],"agent_ref")==0)
		{
			traffic_agent_ref= (TrafficAgent*)Tcl::instance().lookup(argv[2]);
			traffic_agent_ref->address= ra_addr_;

			return TCL_OK;

			//The number of nodes used by the simulation
		} else if(strcmp(argv[1],"n_nodes")==0)
		{
			total_nodes_= atoi(argv[2]);
			cache_= (flatbeacon_cache*) new flatbeacon_cache(total_nodes_);

			return TCL_OK;
		}
	}
	// Pass the command to the base class
	return Agent::command(argc, argv);
}



/// This function is called whenever a packet is received. It identifies
/// the type of the received packet and process it accordingly.
void flatbeacon::recv(Packet* p, Handler* h) {
	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_ip* ih = HDR_IP(p);

	// Access the Ping header for the received packet:

	if(is_online)
	{
		if (ih->saddr() == ra_addr()) {
			// If there exists a loop, must drop the packet
			if (ch->num_forwards() > 0) {
				drop(p, DROP_RTR_ROUTE_LOOP);
				return;
			}
			// else if this is a packet I am originating, must add IP header
			else if (ch->num_forwards() == 0)
				ch->size() += IP_HDR_LEN;
		}

		// If it is a flatbeacon packet, must process it
		if (ch->ptype() == PT_FLATBEACON)
			recv_flatbeacon_pkt(p);
		// Otherwise, must forward the packet (unless TTL has reached zero)
		else {
			ih->ttl_--;
			if (ih->ttl_ == 0) {
				drop(p, DROP_RTR_TTL);
				return;
			}
			forward_data(p);
		}
	} else
		Packet::free(p);
}


void flatbeacon::recv_flatbeacon_pkt(Packet* p) {
	struct hdr_ip* ih = HDR_IP(p);

	// All routing messages are sent from and to port RT_PORT,
	// so we check it.
	assert(ih->sport() == RT_PORT);
	assert(ih->dport() == RT_PORT);

	/* ... processing of flatbeacon packet ... */
	recv_p2p_pkt(p);

	// Release resources
	Packet::free(p);
}



///This function sees if it is any bucket below K and sends a query to the BGL or to all
///neighbours

void flatbeacon::send_query(){

	std::vector<int> buck;


	//is_blg_queried = 0 -> BGL
	//is_blg_queried != 0 -> all the neighbours

	if(neigh_table_->buck_query(K_,codebit_,&buck))
	{

		if(tab_bc->bc_gl_addr!=-1 && tab_bc->bc_gl_addr!= ra_addr_  && is_bgl_queried ==0){



			send_msg_query(buck,tab_bc->bc_gl_addr, tab_bc->bc_gl_addr);


		} else {



			if(is_bgl_queried==0 || is_bgl_queried==2){

				is_bgl_queried=0;

				for(int i=0;i<codebit_;i++)
				{
					for(array_nodes::iterator it= neigh_table_->bk_[i].begin();it!= neigh_table_->bk_[i].end();it++)
					{
						flatbeacon_neib_node* nb_node= *it;
						if(nb_node->dest_addr_!= tab_bc->bc_gl_addr && nb_node->num_hops_< HOP_QUERY)
						{


							send_msg_query(buck,nb_node->dest_addr_,nb_node->next_hop_addr_);
						}


					}
				}
			}
		}
	}
}


//Sends QUERY MESSAGE

void flatbeacon::send_msg_query(std::vector<int> buck,nsaddr_t dest, nsaddr_t next_hop)
{

	int id= cache_->update_query_cache_control(dest);

	cache_->update_query_cache(dest,ra_addr_,id);


	Packet* p= allocpkt();
	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_ip* ih = HDR_IP(p);
	struct hdr_flatbeacon_pkt* ph = HDR_FLATBEACON_PKT(p);
	struct hdr_flatbeacon_query_pkt* pq= HDR_FLATBEACON_QUERY_PKT(p);
	ch->ptype() = PT_FLATBEACON;

	ch->direction() = hdr_cmn::DOWN;

	ch->error() = 0;
	ch->next_hop() = next_hop;
	ch->addr_type() = NS_AF_INET;
	if (use_mac()) {
		ch->xmit_failure_ = flatbeacon_mac_failed_callback;
		ch->xmit_failure_data_ = (void*) this;
	}
	ih->saddr() = ra_addr();
	ih->daddr() = next_hop;
	ih->sport() = RT_PORT;
	ih->dport() = RT_PORT;
	ih->ttl() = IP_DEF_TTL;

	pq->qry_tyre() = PT_QRY;
	pq->buckets_= buck;
	//pq->fil_= b_filter;
	pq->pkt_src()= ra_addr_;
	pq->pkt_hop()= ra_addr_;
	pq->pkt_dest()= dest;
	pq->num_hop()= 1;
	pq->id_= id;

	ch->size() = IP_HDR_LEN + ph->pkt_len()+pq->size();

	Scheduler::instance().schedule(target_, p, JITTER);
}




///Sends a flatbeacon_pkt (HELLO)
void flatbeacon::send_flatbeacon_pkt() {
	Packet* p = allocpkt();
	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_ip* ih = HDR_IP(p);
	struct hdr_flatbeacon_pkt* ph = HDR_FLATBEACON_PKT(p);
	struct hdr_p2p_bc *bc = HDR_P2P_BC(p);

	ph->pkt_src() = ra_addr();
	ph->pkt_len() = 7;
	ph->pkt_seq_num() = seq_num_;



	//BEACONS
	bc->bc_type = P2P_PT_BC;
	bc->bc_seq_num = seq_num_;
	seq_num_++;
	bc->bc_associativity = tab_bc->bc_associativity;
	bc->bc_gl_addr = tab_bc->bc_gl_addr;
	//////


	ch->ptype() = PT_FLATBEACON;

	ch->direction() = hdr_cmn::DOWN;
	ch->size() = IP_HDR_LEN + ph->pkt_len() + bc->size();
	ch->error() = 0;
	ch->next_hop() = IP_BROADCAST;
	ch->addr_type() = NS_AF_INET;
	if (use_mac()) {
		ch->xmit_failure_ = flatbeacon_mac_failed_callback;
		ch->xmit_failure_data_ = (void*) this;
	}
	ih->saddr() = ra_addr();
	ih->daddr() = IP_BROADCAST;
	ih->sport() = RT_PORT;
	ih->dport() = RT_PORT;
	ih->ttl() = 1;
	Scheduler::instance().schedule(target_, p, 0);
}

void flatbeacon::reset_flatbeacon_pkt_timer() {
	double interval = BEACON_AT_INTERVAL + BEACON_P2P_JITTER;
	pkt_timer_.resched(interval);
}

void flatbeacon::reset_timer_query()
{
	double interval = QUERY_TIMEOUT;
	pkt_timer_qry_.resched(interval);
}

void flatbeacon::reset_timer_send_bgl()
{
	double interval= SEND_BGL_TIMEOUT;
	pkt_timer_send_bgl_timeout_.resched(interval);
}


///Forwards a data packet to the appropiate next hop indicated by the routing table.
void flatbeacon::forward_data(Packet* p) {
	struct hdr_cmn* ch = HDR_CMN(p);
	struct hdr_ip* ih = HDR_IP(p);

	neigh_table_->update_timer_neib(ih->saddr(),codebit_);

	if (ch->direction() == hdr_cmn::UP && ((u_int32_t) ih->daddr()
			== IP_BROADCAST || ih->daddr() == ra_addr())) {
		dmux_->recv(p, 0);
		return;
	} else {
		ch->direction() = hdr_cmn::DOWN;
		ch->addr_type() = NS_AF_INET;
		if ((u_int32_t) ih->daddr() == IP_BROADCAST)
			ch->next_hop() = IP_BROADCAST;
		else {

			nsaddr_t  next_hop;
			//rtable_.print(logtarget_);

			int result_xor= ih->daddr() ^ ra_addr_;
			int for_bucket=find_common_prefix(result_xor);


			////////// 1 RULE/////// ENGLOBES THE DROP//////////////

			std::map<flatbeacon_neib_node*,int> prefix;
			prefix.clear();


			for(array_nodes::iterator it=neigh_table_->bk_[for_bucket].begin();it!=neigh_table_->bk_[for_bucket].end();it++)
			{
				flatbeacon_neib_node* nb_node = *it;

				int result_prefix= find_full_common_prefix(ih->daddr(),nb_node->dest_addr_);

				if(prefix.size() != 0)
				{

					for(std::map<flatbeacon_neib_node*,int>::iterator it2= prefix.begin(); it2 != prefix.end(); it2++)
					{
						int map_pre= (*it2).second;

						if(result_prefix > map_pre)
						{
							prefix.clear();
							prefix.insert(pair<flatbeacon_neib_node*,int>(nb_node,result_prefix));
							break;
						} else if(result_prefix == map_pre )
						{
							prefix.insert(pair<flatbeacon_neib_node*,int>(nb_node,result_prefix));
							break;
						} else
						{
							break;
						}
					}

				} else {

					prefix.insert(pair<flatbeacon_neib_node*,int>(nb_node,result_prefix));
				}
			}


			if(prefix.size() == 1)
			{
				std::map<flatbeacon_neib_node*,int>::iterator it2= prefix.begin();
				next_hop= (*it2).first->next_hop_addr_;

			} else if(prefix.size() == 0)
			{

					//////DROP//////////


				drop(p, DROP_RTR_NO_ROUTE);
				return;

			} else {


				///// 2 RULE/////////////


				std::vector<flatbeacon_neib_node*> hops_prefix;
				hops_prefix.clear();
				int max_hops= 1000;


				for(std::map<flatbeacon_neib_node*,int>::iterator it2= prefix.begin(); it2 != prefix.end(); it2++)
				{
					flatbeacon_neib_node* num_hop_pre= (*it2).first;

					if(hops_prefix.size()!=0)
					{
						if(num_hop_pre->num_hops_ < max_hops)
						{
							hops_prefix.clear();
							hops_prefix.push_back(num_hop_pre);
							max_hops= num_hop_pre->num_hops_;

						} else if(num_hop_pre->num_hops_ == max_hops)
						{
							hops_prefix.push_back(num_hop_pre);
						}


					}else {

						hops_prefix.push_back(num_hop_pre);
						max_hops= num_hop_pre->num_hops_;
					}

				}

				if(hops_prefix.size()==1)
				{
					for(std::vector<flatbeacon_neib_node*>::iterator it3= hops_prefix.begin(); it3!= hops_prefix.end();it3++)
					{
						flatbeacon_neib_node* nb_max_hops= *it3;
						next_hop= nb_max_hops->next_hop_addr_;
					}
				} else {

					//////// 3 RULE/////////////////


					int max_addr= 500;
					for(std::vector<flatbeacon_neib_node*>::iterator it3= hops_prefix.begin(); it3!= hops_prefix.end();it3++)
					{
						flatbeacon_neib_node* nb_max_hops= *it3;
						if(nb_max_hops->dest_addr_ < max_addr)
						{
							max_addr= nb_max_hops->dest_addr_;
							next_hop= nb_max_hops->next_hop_addr_;
						}
					}

				}



			}

			if (next_hop == IP_BROADCAST) {
				debug("%f: Agent %d can not forward a packet destined to %d\n",
						CURRENT_TIME, ra_addr(), ih->daddr());
				drop(p, DROP_RTR_NO_ROUTE);
				return;
			} else{

				pathTableType test;
				if (ch->uid_ != 0)
					test.add_path_table(ra_addr(), p, ih->saddr(),ih->daddr(), traffic_agent_ref, false);
				ch->next_hop() = next_hop;
			}
		}

		Scheduler::instance().schedule(target_, p, 0.0);
	}
}




///This function processes an incoming flatbeacon packet.

void flatbeacon::recv_p2p_pkt(Packet* pkt) {

	// Access the IP header for the received packet:
	hdr_ip* hdrip = hdr_ip::access(pkt);

	// check if in brdcast mode

	if ((hdrip->saddr() != ra_addr_))
	{

		hdr_p2p_bc* hdr_bc = HDR_P2P_BC(pkt);
		if (hdr_bc->bc_type == P2P_PT_BC) {

			tab_bc->Add_beacon(hdr_bc, &hdrip->saddr());

			bool for_query= false;

			if(hdr_bc->bc_gl_addr == hdrip->saddr())
			{
				for_query= true;
			}

			update_physical_neighbors( hdrip->saddr(),for_query);
			neigh_table_->update_timer_neib(hdrip->saddr(),codebit_);

		}

		if(hdrip->daddr()==ra_addr_)
		{
			hdr_flatbeacon_query_pkt* hdr_qry= HDR_FLATBEACON_QUERY_PKT(pkt);


			if(hdr_qry->qry_type_== PT_QRY)
			{
				cache_->update_query_cache(hdr_qry->pkt_dest_,hdr_qry->pkt_src_,hdr_qry->id_);


				if(hdr_qry->pkt_dest()== ra_addr_)
				{
					neigh_table_->add_entry(hdr_qry->pkt_src_,hdr_qry->pkt_hop_,ra_addr_,codebit_, hdr_qry->num_hop_,
							MSG_QUERY,ra_addr_,hdr_qry->pkt_src_,hdr_qry->id_,false);

					search_neig(pkt,ra_addr_);

					neigh_table_->update_timer_neib(hdrip->saddr(),codebit_);
					neigh_table_->update_timer_neib(hdr_qry->pkt_src_,codebit_);

				} else
				{
					neigh_table_->add_entry(hdr_qry->pkt_src_,hdr_qry->pkt_hop_,ra_addr_,codebit_, hdr_qry->num_hop_,
							MSG_QUERY,hdr_qry->pkt_dest_,hdr_qry->pkt_src_,hdr_qry->id_,false);

					neigh_table_->update_timer_neib(hdr_qry->pkt_src_,codebit_);

					neigh_table_->update_timer_neib(hdrip->saddr(),codebit_);


					int resul_xor= hdr_qry->pkt_dest_ ^ ra_addr_;
					int ind_xor= find_common_prefix(resul_xor);
					for(array_nodes::iterator it= neigh_table_->bk_[ind_xor].begin(); it!= neigh_table_->bk_[ind_xor].end();it++)
					{
						flatbeacon_neib_node* nb_node= *it;
						if(nb_node->dest_addr_==hdr_qry->pkt_dest_)
						{
							Packet* p= allocpkt();
							struct hdr_cmn* ch = HDR_CMN(p);
							struct hdr_ip* ih = HDR_IP(p);
							struct hdr_flatbeacon_pkt* ph = HDR_FLATBEACON_PKT(p);
							struct hdr_flatbeacon_query_pkt* pq= HDR_FLATBEACON_QUERY_PKT(p);
							ch->ptype() = PT_FLATBEACON;

							ch->direction() = hdr_cmn::DOWN;

							ch->error() = 0;
							ch->next_hop() = nb_node->next_hop_addr_;
							ch->addr_type() = NS_AF_INET;
							if (use_mac()) {
								ch->xmit_failure_ = flatbeacon_mac_failed_callback;
								ch->xmit_failure_data_ = (void*) this;
							}
							ih->saddr() = ra_addr();
							ih->daddr() = nb_node->next_hop_addr_;
							ih->sport() = RT_PORT;
							ih->dport() = RT_PORT;
							ih->ttl() = IP_DEF_TTL;

							pq->qry_tyre() = PT_QRY;
							pq->buckets_= hdr_qry->buckets_;
							pq->pkt_src()= hdr_qry->pkt_src_;
							pq->pkt_hop()= ra_addr_;
							pq->pkt_dest()= hdr_qry->pkt_dest_;
							pq->num_hop()= hdr_qry->num_hop_+1;
							pq->id_= hdr_qry->id_;

							ch->size() = IP_HDR_LEN + ph->pkt_len()+pq->size();

							Scheduler::instance().schedule(target_, p, JITTER);
						}
					}
				}

			} else if(hdr_qry->qry_type_== PT_REPLY_QUERY)
			{

				cache_->update_response_cache(hdr_qry->pkt_dest_,hdr_qry->pkt_src_,hdr_qry->id_);

				neigh_table_->add_entry(hdr_qry->pkt_src_,hdr_qry->pkt_hop_,ra_addr_,codebit_, hdr_qry->num_hop_,
						MSG_RESP,hdr_qry->pkt_src_,hdr_qry->pkt_dest_,hdr_qry->id_,false);

				neigh_table_->update_timer_neib(hdrip->saddr(),codebit_);
				neigh_table_->update_timer_neib(hdr_qry->pkt_src_,codebit_);

				if(hdr_qry->pkt_dest_== ra_addr_)
				{
					neigh_table_->insert_neib_query(hdr_qry->pkt_hop_,hdr_qry->buckets_,ra_addr_,codebit_, hdr_qry->buck_hop_,hdr_qry->num_hop_,
							true,MSG_RESP,hdr_qry->pkt_dest_,hdr_qry->pkt_src_,hdr_qry->id_,hdr_qry->line_,hdr_qry->col_,
							hdr_qry->type_,hdr_qry->ids_);
				} else
				{
					neigh_table_->insert_neib_query(hdr_qry->pkt_hop_,hdr_qry->buckets_,ra_addr_,codebit_, hdr_qry->buck_hop_,hdr_qry->num_hop_,
							false,MSG_RESP,hdr_qry->pkt_dest_,hdr_qry->pkt_src_,hdr_qry->id_,hdr_qry->line_,hdr_qry->col_,
							hdr_qry->type_,hdr_qry->ids_);
				}

				if(hdr_qry->pkt_src_==tab_bc->bc_gl_addr && hdr_qry->pkt_dest_== ra_addr_ && is_bgl_queried == 1)
				{
					is_bgl_queried=0;
					//pkt_timer_bgl_timeout_.force_cancel();
				}

				if(hdr_qry->pkt_dest_!= ra_addr_)
				{

					int resul_xor= hdr_qry->pkt_dest_ ^ ra_addr_;

					int ind_xor= find_common_prefix(resul_xor);

					for(array_nodes::iterator it= neigh_table_->bk_[ind_xor].begin(); it!= neigh_table_->bk_[ind_xor].end();it++)
					{
						flatbeacon_neib_node* nb_node= *it;

						if(nb_node->dest_addr_==hdr_qry->pkt_dest_)
						{
							Packet* p= allocpkt();
							struct hdr_cmn* ch = HDR_CMN(p);
							struct hdr_ip* ih = HDR_IP(p);
							struct hdr_flatbeacon_pkt* ph = HDR_FLATBEACON_PKT(p);
							struct hdr_flatbeacon_query_pkt* pq= HDR_FLATBEACON_QUERY_PKT(p);
							ch->ptype() = PT_FLATBEACON;

							ch->direction() = hdr_cmn::DOWN;

							ch->error() = 0;
							ch->next_hop() = nb_node->next_hop_addr_;
							ch->addr_type() = NS_AF_INET;
							if (use_mac()) {
								ch->xmit_failure_ = flatbeacon_mac_failed_callback;
								ch->xmit_failure_data_ = (void*) this;
							}
							ih->saddr() = ra_addr();
							ih->daddr() = nb_node->next_hop_addr_;
							ih->sport() = RT_PORT;
							ih->dport() = RT_PORT;
							ih->ttl() = IP_DEF_TTL;

							pq->qry_tyre() = PT_REPLY_QUERY;
							pq->buckets_= hdr_qry->buckets_;
							pq->pkt_src()= hdr_qry->pkt_src_;
							pq->pkt_hop()= ra_addr_;
							pq->pkt_dest()= hdr_qry->pkt_dest_;
							pq->buck_hop_= hdr_qry->buck_hop_;
							pq->num_hop()= hdr_qry->num_hop_+1;
							pq->id_= hdr_qry->id_;
							pq->col_= hdr_qry->col_;
							pq->line_= hdr_qry->line_;
							pq->type_= hdr_qry->type_;
							pq->ids_= hdr_qry->ids_;

							ch->size() = IP_HDR_LEN + ph->pkt_len()+pq->size();

							Scheduler::instance().schedule(target_, p, 	JITTER);
						}
					}

				}
			} else if(hdr_qry->qry_type_ == PT_TAB_BGL){


				neigh_table_->insert_neib_bgl(hdr_qry->pkt_hop_,hdr_qry->buckets_,ra_addr_,codebit_,hdr_qry->buck_hop_,
						hdr_qry->num_hop_,hdr_qry->line_,hdr_qry->col_,hdr_qry->type_,hdr_qry->ids_);

				neigh_table_->update_timer_neib(hdr_qry->pkt_src_,codebit_);
			}
		}

		Packet::free(pkt);
		return;
	}
}


///Update the routing table when a flatbeacon_pkt is received

void flatbeacon::update_physical_neighbors(nsaddr_t src, bool for_query)
{

	neigh_table_->add_entry(src,src,ra_addr_,codebit_,1,MSG_HELLO,ra_addr_,src,-1,for_query);

}


///Sends a RESPONSE when it's has nodes for the node which perfoms the query.

void flatbeacon::search_neig(Packet* pkt, nsaddr_t orig)
{

	std::vector<int> find_neig;
	std::vector<int> bucks;
	std::vector<int> buck_hop;
	std::vector<int> lines;
	std::vector<int> cols;
	std::vector<int> types;
	std::vector<int> ids;
	int count_buck;


	hdr_flatbeacon_query_pkt* hdr_qry= HDR_FLATBEACON_QUERY_PKT(pkt);
	hdr_ip* hdr_ih = HDR_IP(pkt);
	bucks= hdr_qry->buckets_;

	if(neigh_table_->search_nodes_bucks(hdr_qry->pkt_src_,bucks,codebit_,&find_neig,&count_buck,&buck_hop,&cols,
			&lines,&types,&ids))
	{

		int id= cache_->update_response_cache_control(hdr_qry->pkt_src_);

		cache_->update_response_cache(hdr_qry->pkt_src_, ra_addr_,id);

		Packet* p= allocpkt();
		struct hdr_cmn* ch = HDR_CMN(p);
		struct hdr_ip* ih = HDR_IP(p);
		struct hdr_flatbeacon_pkt* ph = HDR_FLATBEACON_PKT(p);
		struct hdr_flatbeacon_query_pkt* pq= HDR_FLATBEACON_QUERY_PKT(p);
		ch->ptype() = PT_FLATBEACON;

		ch->direction() = hdr_cmn::DOWN;

		ch->error() = 0;
		ch->next_hop() = hdr_ih->saddr();
		ch->addr_type() = NS_AF_INET;
		if (use_mac()) {
			ch->xmit_failure_ = flatbeacon_mac_failed_callback;
			ch->xmit_failure_data_ = (void*) this;
		}
		ih->saddr() = ra_addr();
		ih->daddr() = hdr_ih->saddr();
		ih->sport() = RT_PORT;
		ih->dport() = RT_PORT;
		ih->ttl() = IP_DEF_TTL;

		pq->qry_tyre() = PT_REPLY_QUERY;
		pq->buckets_= find_neig;
		pq->pkt_src()= ra_addr_;
		pq->pkt_hop()= ra_addr_;
		pq->pkt_dest()= hdr_qry->pkt_src_;
		pq->num_hop()= 1;
		pq->buck_hop_= buck_hop;
		pq->id_= id;
		pq->line_= lines;
		pq->col_= cols;
		pq->type_= types;
		pq->ids_= ids;

		ch->size() = IP_HDR_LEN + ph->pkt_len()+pq->size();


		Scheduler::instance().schedule(target_, p, JITTER);

	}


}


///find the common prefix

int flatbeacon::find_common_prefix(int result_xor)
{
	int mask,k;
	char node_xor[codebit_];
	for( int t =(codebit_-1) ; t >= 0 ; t--)
	{
		 mask = 1 << t;
		 k = result_xor & mask;
		 if( k == 0)
			 node_xor[t]='0';
		 else
			 node_xor[t]='1';
	}

	int count_xor=0;
	for(int y=(codebit_-1);y>=0;y--)
	{
		if(node_xor[y]=='0')
		{
			count_xor++;
		} else
			break;
	}

	return (codebit_-1)-count_xor;
}


///This function returns the number os bits which are diferente between tho numbers

int flatbeacon::find_full_common_prefix(nsaddr_t forwa,nsaddr_t orig)
{
	char node_dest[codebit_];
	char node_qry[codebit_];

	int i  , k ,k1, mask;

	for( i =(codebit_-1) ; i >= 0 ; i--)
	{
		 mask = 1 << i;
		 k = forwa & mask;
		 k1= orig & mask;
		 if( k == 0)
		   node_qry[i]='0';
		 else
			node_qry[i]='1';

		 if(k1==0)
			 node_dest[i]='0';
		 else
			 node_dest[i]='1';
	}

	int count=0;
	for(int i=0; i< codebit_; i++)
	{
		if(node_dest[i] == node_qry[i])
			count++;
	}

	return count;

}


///This fucntion sends all the informantion contained in the routing table
///to the BGL.

void flatbeacon::update_bgl(){

	if(tab_bc->bc_gl_addr == -1)
	{
		BGL= -1;
		pkt_timer_send_bgl_timeout_.force_cancel();

	}else if(BGL != tab_bc->bc_gl_addr && tab_bc->bc_gl_addr != ra_addr_ )
	{


		double interval= SEND_BGL_TIMEOUT;

		pkt_timer_send_bgl_timeout_.force_cancel();
		pkt_timer_send_bgl_timeout_.sched(interval);


		BGL= tab_bc->bc_gl_addr;

		send_bgl();
	}
}


///Sends the message which contains the routing table.

void flatbeacon::send_bgl()
{

	if(tab_bc->bc_gl_addr != -1 && tab_bc->bc_gl_addr != ra_addr_)
	{

		std::vector<int> nodes;
		std::vector<int> hops_nodes;
		std::vector<int> cols;
		std::vector<int> lines;
		std::vector<int> ids;
		std::vector<int> types;

		nodes.clear();
		hops_nodes.clear();
		cols.clear();
		lines.clear();
		ids.clear();
		types.clear();

		for(int i= 0; i< codebit_; i++)
		{
			for(array_nodes::iterator it= neigh_table_->bk_[i].begin(); it!= neigh_table_->bk_[i].end();it++)
			{
				flatbeacon_neib_node* nb_node= *it;

				nodes.push_back(nb_node->dest_addr_);
				hops_nodes.push_back(nb_node->num_hops_);
				cols.push_back(nb_node->col_);
				lines.push_back(nb_node->line_);
				ids.push_back(nb_node->id_);
				types.push_back(nb_node->type_msg_);
			}
		}


		Packet* p= allocpkt();
		struct hdr_cmn* ch = HDR_CMN(p);
		struct hdr_ip* ih = HDR_IP(p);
		struct hdr_flatbeacon_pkt* ph = HDR_FLATBEACON_PKT(p);
		struct hdr_flatbeacon_query_pkt* pq= HDR_FLATBEACON_QUERY_PKT(p);
		ch->ptype() = PT_FLATBEACON;

		ch->direction() = hdr_cmn::DOWN;

		ch->error() = 0;
		ch->next_hop() = tab_bc->bc_gl_addr;
		ch->addr_type() = NS_AF_INET;
		if (use_mac()) {
			ch->xmit_failure_ = flatbeacon_mac_failed_callback;
			ch->xmit_failure_data_ = (void*) this;
		}
		ih->saddr() = ra_addr();
		ih->daddr() = tab_bc->bc_gl_addr;
		ih->sport() = RT_PORT;
		ih->dport() = RT_PORT;
		ih->ttl() = IP_DEF_TTL;

		pq->qry_tyre() = PT_TAB_BGL;
		//printf("\nquery_buckets NODE: %d  bucks: %d\n",ra_addr_,buck);
		pq->buckets_= nodes;
		pq->buck_hop_= hops_nodes;
		pq->col_= cols;
		pq->line_= lines;
		pq->ids_= ids;
		pq->type_= types;
		pq->pkt_src()= ra_addr_;
		pq->pkt_hop()= ra_addr_;
		pq->pkt_dest()= tab_bc->bc_gl_addr;
		pq->num_hop()= 1;
		pq->id_= -1;


		ch->size() = IP_HDR_LEN + ph->pkt_len()+pq->size();

		Scheduler::instance().schedule(target_, p, 0);
	}
}


///This functios sees if exists any bucktes below K.

void flatbeacon::send_query_new(nsaddr_t dest, nsaddr_t next_hop)
{
	std::vector<int> buck;
	if(neigh_table_->buck_query(K_,codebit_,&buck))
	{
		send_msg_query(buck,dest,next_hop);
	}
}



