/*********************************************************************************

 ** Universidade Nova de Lisboa - Faculdade de Ciências e Tecnologia

 **

 ** Protocolo XORi - flatbeacon.h

 **

 ** Autores:

 ** André Garrido

 ** Rodolfo Oliveira

 **

 ** 1 de Março de 2010

** **********************************************************************************/



#ifndef __flatbeacon_h__
#define __flatbeacon_h__
#include  "flatbeacon_pkt.h"
#include  "flatbeacon_query_pkt.h"
#include  "flatbeacon_rtable.h"
#include  "../r0_clustering/cbc_table.h"
#include "flatbeacon_repositories.h"
#include "flatbeacon_neigb_table.h"
#include  <agent.h>
#include  <packet.h>
#include  <string>
#include  <trace.h>
#include  <timer-handler.h>
#include  <random.h>
#include  <rng.h>
#include  <classifier.h>
#include  <map>
#include "FCT_traffic_app/traffic_agent.h"
#include "flatbeacon_cache.h"



#define CURRENT_TIME          Scheduler::instance().clock()
#define JITTER                (Random::uniform()*0.5)
#define BGL_RES_TIMEOUT		1
#define QUERY_TIMEOUT 5     //Query messages emission interval
#define SEND_BGL_TIMEOUT 4  //BGL routing table messages emission interval
#define HOP_QUERY	5       //Number of hops in which the query is limited

enum flatbeacon_state {
	vazio, cheio
};

class flatbeacon; // forward declaration
class Beacons_tab;
class flatbeacon_neigb_table;
class flatbeacon_cache;
/* Timers */
class flatbeacon_PktTimer: public TimerHandler {
public:
	flatbeacon_PktTimer(flatbeacon* agent) :
		TimerHandler() {
		agent_ = agent;
	}
protected:
	flatbeacon* agent_;
	virtual void expire(Event* e);
};

class flatbeacon_PktTimer_query: public TimerHandler{
public:
	flatbeacon_PktTimer_query(flatbeacon* agent):
		TimerHandler(){
		agent_=agent;
	}
protected:
	flatbeacon* agent_;
	virtual void expire(Event* e);
};

class flatbeacon_PktTimer_bgl_timeout: public TimerHandler{
public:
	flatbeacon_PktTimer_bgl_timeout(flatbeacon* agent):
		TimerHandler(){
		agent_=agent;
	}
protected:
	flatbeacon* agent_;
	virtual void expire(Event* e);
};

class flatbeacon_PktTimer_send_bgl_timeout: public TimerHandler{
public:
	flatbeacon_PktTimer_send_bgl_timeout(flatbeacon* agent):
		TimerHandler(){
		agent_= agent;
	}
protected:
	flatbeacon* agent_;
	virtual void expire(Event* e);
};


/* Agent */
class flatbeacon: public Agent {
	/* Friends */
	friend class flatbeacon_PktTimer;
	friend class flatbeacon_PktTimer_query;
	friend class flatbeacon_PktTimer_bgl_timeout;
	friend class flatbeacon_PktTimer_send_bgl_timeout;

	friend class flatbeacon_neigb_table;
	friend class flatbeacon_cache;

	/* Private members  */
	nsaddr_t ra_addr_;
	flatbeacon_state state_;
	flatbeacon_rtable rtable_;
	flatbeacon_neigb_table* neigh_table_;
	flatbeacon_cache* cache_;
	u_int8_t seq_num_;

	//SendBeacon_Timer bc_send_timer;
	Beacons_tab *tab_bc;
	/// Determines if layer 2 notifications are enabled or not.
	int		use_mac_;

	int K_;				//Value of K
	int is_bgl_queried;		// State for the BGL query
	int codebit_;			// number of bits
	bool is_online;
	nsaddr_t BGL;		//Address of the BGL
	int total_nodes_;	// total of nodes used in the simulation



	TrafficAgent* traffic_agent_ref;



	int accessible_var_;
protected:
	Classifier* dmux_; // For passing packets up to agents.
	Trace* logtarget_; // For logging.
	flatbeacon_PktTimer pkt_timer_; // Timer for sending packets.
	flatbeacon_PktTimer_query pkt_timer_qry_;
	flatbeacon_PktTimer_bgl_timeout pkt_timer_bgl_timeout_;
	flatbeacon_PktTimer_send_bgl_timeout pkt_timer_send_bgl_timeout_;
	inline nsaddr_t& ra_addr() {
		return ra_addr_;
	}
	inline flatbeacon_state& state() {
		return state_;
	}
	inline int& accessible_var() {
		return this->accessible_var_;
	}

	inline int&		use_mac()	{ return use_mac_; }

	void forward_data(Packet*);
	void recv_flatbeacon_pkt(Packet*);
	void send_flatbeacon_pkt();
	void send_query();
	void send_msg_query(std::vector<int>, nsaddr_t, nsaddr_t);
	void send_bgl();

	void reset_flatbeacon_pkt_timer();
	void reset_timer_query();
	void reset_timer_send_bgl();

	void recv_p2p_pkt(Packet*);
	void update_physical_neighbors(nsaddr_t src, bool for_query);
	void search_neig(Packet*,nsaddr_t);
	int find_common_prefix(int);
	int find_full_common_prefix(nsaddr_t,nsaddr_t);

	void send_query_new(nsaddr_t,nsaddr_t);



public:


	flatbeacon(nsaddr_t);
	~flatbeacon();
	int command(int, const char* const *);
	void recv(Packet*, Handler*);
	void mac_failed(Packet*);
	void update_bgl();


};
#endif

