/*********************************************************************************

 ** Universidade Nova de Lisboa - Faculdade de Ciências e Tecnologia

 **

 ** Protocolo XOR - flatrouting.h

 **

 ** Autores:

 ** André Garrido

 ** Rodolfo Oliveira

 **

 ** 1 de Março de 2010

** **********************************************************************************/


#ifndef __flatrouting_h__
#define __flatrouting_h__
#include  "flatrouting_pkt.h"
#include  "flat_query_pkt.h"
#include  "flatrouting_rtable.h"
#include  "p2p_packets.h"
#include "flatrouting_repositories.h"
#include "flatrouting_neigb_table.h"
#include  <agent.h>
#include  <packet.h>
#include <string>
#include  <trace.h>
#include  <timer-handler.h>
#include  <random.h>
#include  <classifier.h>
#include <map>
#include "FCT_traffic_app/traffic_agent.h"
#include "flatrouting_cache.h"
#define CURRENT_TIME          Scheduler::instance().clock()
#define JITTER                (Random::uniform()*0.5)
#define BGL_RES_TIMEOUT		1
#define QUERY_TIMEOUT 5     //Query messages emission interval
#define HOP_QUERY	5       //Number of hops in which the query is limited


#define MAJORANTE_JITTER 0.1
#define BEACON_P2P_JITTER (MAJORANTE_JITTER * Random::uniform())
#define BEACON_AT_INTERVAL	         0.95



enum flatrouting_state {
	vazio, cheio
};

class Flatrouting; // forward declaration
class flatrouting_neigb_table;
class flatrouting_cache;


/* Timers */
class Flatrouting_PktTimer: public TimerHandler {
public:
	Flatrouting_PktTimer(Flatrouting* agent) :
		TimerHandler() {
		agent_ = agent;
	}
protected:
	Flatrouting* agent_;
	virtual void expire(Event* e);
};

class Flatrouting_PktTimer_query: public TimerHandler{
public:
	Flatrouting_PktTimer_query(Flatrouting* agent):
		TimerHandler(){
		agent_=agent;
	}
protected:
	Flatrouting* agent_;
	virtual void expire(Event* e);
};


/* Agent */
class Flatrouting: public Agent {
	/* Friends */
	friend class Flatrouting_PktTimer;
	friend class Flatrouting_PktTimer_query;

	friend class flatrouting_neigb_table;
	friend class flatrouting_cache;
	/* Private members  */
	nsaddr_t ra_addr_;
	flatrouting_state state_;
	flatrouting_rtable rtable_;
	flatrouting_neigb_table* neigh_table_;
	flatrouting_cache* cache_;
	u_int8_t seq_num_;

	/// Determines if layer 2 notifications are enabled or not.
	int		use_mac_;


	int codebit_;   // number of bits
	int K_;         //Value of K
	bool is_online;
	TrafficAgent* traffic_agent_ref;

	int total_nodes_;    // total of nodes used in the simulation



	int accessible_var_;
protected:
	Classifier* dmux_; // For passing packets up to agents.
	Trace* logtarget_; // For logging.
	Flatrouting_PktTimer pkt_timer_; // Timer for sending packets.
	Flatrouting_PktTimer_query pkt_timer_qry_;
	inline nsaddr_t& ra_addr() {
		return ra_addr_;
	}
	inline flatrouting_state& state() {
		return state_;
	}
	inline int& accessible_var() {
		return this->accessible_var_;
	}

	inline int&		use_mac()	{ return use_mac_; }

	void forward_data(Packet*);
	void recv_flatrouting_pkt(Packet*);
	void send_flatrouting_pkt();
	void send_query();
	void send_msg_query(std::vector<int>, nsaddr_t, nsaddr_t);
	void reset_flatrouting_pkt_timer();
	void reset_timer_query();

	void recv_p2p_pkt(Packet*);
	void update_physical_neighbors(nsaddr_t src);
	void search_neig(Packet*,nsaddr_t);
	int find_common_prefix(int);
	int find_full_common_prefix(nsaddr_t,nsaddr_t);

	void send_query_new(nsaddr_t,nsaddr_t);



public:
	Flatrouting(nsaddr_t);
	~Flatrouting();
	int command(int, const char* const *);
	void recv(Packet*, Handler*);
	void mac_failed(Packet*);

	int is_bgl_queried;
};
#endif

